import { INDENT, INTERNAL, formatType, toHex } from './common';

export class Method {
    readonly self: string;
    readonly shortName: string;
    readonly returnType: string;
    readonly args: string;
    readonly address: number;
    readonly isInherited: boolean;
    readonly isStatic: boolean;

    // Constructor
    constructor(self: string, name: string, returnType: string, args: string, address: number, isInherited: boolean, isStatic: boolean) {
        this.self = self;
        this.shortName = name;
        this.returnType = returnType;
        this.args = args;
        this.address = address;
        this.isInherited = isInherited;
        this.isStatic = isStatic;
    }

    // Getters
    getName(separator = '_') {
        return this.self + separator + this.shortName;
    }
    getType() {
        return this.getName() + '_t';
    }
    #getRawType() {
        return INTERNAL + 'raw_' + this.getType();
    }
    getProperty() {
        return `${INDENT}${this.#getRawType()} *${this.shortName};\n`;
    }
    #getFullType() {
        return `${INTERNAL}Function<${this.#getRawType()}>`;
    }

    // Typedefs
    generateTypedefs() {
        let out = '';
        out += `typedef ${formatType(this.returnType.trim())}${this.#getRawType()}${this.args.trim()};\n`;
        out += `typedef const std::function<${this.#getRawType()}> &${this.getType()};\n`;
        return out;
    }

    // Overwrite Helper
    #getVirtualCall(self: string = this.self) {
        return `${self}_vtable::base->${this.shortName}`;
    }
    generate(code: boolean, isVirtual: boolean, parentSelf?: string) {
        let out = '';
        out += 'extern ';
        const type = this.#getFullType();
        out += `${type} *const ${this.getName()}`;
        if (code) {
            out += ` = new ${type}(${JSON.stringify(this.getName('::'))}, `;
            out += `${INTERNAL}thunk<&${this.getName()}>, `;
            if (isVirtual) {
                const parentMethod = parentSelf ? this.#getVirtualCall(parentSelf) : 'nullptr';
                out += `&${this.#getVirtualCall()}, ${parentMethod}`;
            } else {
                out += `(${this.#getRawType()} *) ${toHex(this.address)}`;
            }
            out += ')';
        }
        out += ';\n';
        return out;
    }
}